/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 82-83.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.*;

import org.joml.*;

import java.nio.*;
import javax.swing.*;
import java.lang.Math;

public class Code_12 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[2];
   private float cameraX, cameraY, cameraZ;
   private float cubeLocX, cubeLocY, cubeLocZ;
   private int renderingProgram;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16); // for transfering matrix to uniform
   private final Matrix4f pMat = new Matrix4f();  // perspective matrix
   private final Matrix4f vMat = new Matrix4f();  // view matrix
   private final Matrix4f mMat = new Matrix4f();  // model matrix
   private final Matrix4f mvMat = new Matrix4f(); // model-view matrix
   private int vLoc, projLoc, tfLoc;
   private float aspect;
   private double timeFactor;
   private double startTime;
   private double elapsedTime;
   private double whichCube;

   public Code_12()
   {
      this.setTitle("Chapter 4 - program 2 (Code_12)");
      this.setSize(600, 600);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);

      final Animator animator = new Animator(myCanvas);
      animator.start();
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_DEPTH_BUFFER_BIT);
      gl.glClear(GL_COLOR_BUFFER_BIT);

      gl.glUseProgram(renderingProgram);

         vLoc = gl.glGetUniformLocation(renderingProgram, "v_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.setPerspective((float)Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);

      vMat.translation(-cameraX, -cameraY, -cameraZ);

      elapsedTime = System.currentTimeMillis() - startTime;
      timeFactor = elapsedTime/10000.0; // time factor = 1000 for 24 instances, 10000 for 100000 instances

      gl.glUniformMatrix4fv(   vLoc, 1, false, vMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false, pMat.get(vals));

      tfLoc = gl.glGetUniformLocation(renderingProgram, "tf");
      gl.glUniform1f(tfLoc, (float)timeFactor);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glEnable(GL_DEPTH_TEST);
      gl.glDepthFunc(GL_LEQUAL);

      gl.glDrawArraysInstanced(GL_TRIANGLES, 0, 36, 100000); // 0, 36, 24     when 24 instances
                                                             // 0, 36, 100000 when 100000 instances
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) drawable.getGL();

      startTime = System.currentTimeMillis();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_12.glsl",
                                                   "shaders/fragShader_9.glsl");
      setupVertices();

       cameraX = 0.0f;  cameraY =  0.0f;  cameraZ = 420.0f;
      cubeLocX = 0.0f; cubeLocY = -2.0f; cubeLocZ = 0.0f;
   }

   private void setupVertices()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final float[] vertexPositions =
      {
        -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,
        -1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,
         1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f, -1.0f
      };

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(vbo.length, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer vertBuf = Buffers.newDirectFloatBuffer(vertexPositions);
      gl.glBufferData(GL_ARRAY_BUFFER, vertBuf.limit()*4, vertBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_12();
   }
}
