/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 138-140.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.texture.*;

import org.joml.*;

import java.io.*;
import java.nio.*;
import java.lang.Math;
import javax.swing.*;

public class Code_17 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[3];
   private float cameraX, cameraY, cameraZ;
   private float sphLocX, sphLocY, sphLocZ;
   private int renderingProgram;
   private int earthTexture;

   private Sphere mySphere;
   private int numSphereVerts;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16);
   private final Matrix4f pMat = new Matrix4f();  // perspective matrix
   private final Matrix4f vMat = new Matrix4f();  // view matrix
   private final Matrix4f mMat = new Matrix4f();  // model matrix
   private final Matrix4f mvMat = new Matrix4f(); // model-view matrix
   private int mvLoc, projLoc;
   private float aspect;

   public Code_17()
   {
      this.setTitle("Chapter 6 - program 1 (Code_17)");
      this.setSize(800, 800);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClear(GL_DEPTH_BUFFER_BIT);

      gl.glUseProgram(renderingProgram);

        mvLoc = gl.glGetUniformLocation(renderingProgram, "mv_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      vMat.identity().setTranslation(-cameraX, -cameraY, -cameraZ);

      mMat.identity();
      mMat.translate(sphLocX, sphLocY, sphLocZ);

      mvMat.identity();
      mvMat.mul(vMat);
      mvMat.mul(mMat);

      gl.glUniformMatrix4fv(  mvLoc, 1, false, mvMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false,  pMat.get(vals));

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      gl.glVertexAttribPointer(1, 2, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(1);

      gl.glActiveTexture(GL_TEXTURE0);
      gl.glBindTexture(GL_TEXTURE_2D, earthTexture);

      gl.glEnable(GL_CULL_FACE);
      gl.glFrontFace(GL_CCW);
      gl.glDrawArrays(GL_TRIANGLES, 0, numSphereVerts);
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_17.glsl",
                                                   "shaders/fragShader_17.glsl");

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);

      setupVertices();

      cameraX = 0.0f; cameraY = 0.0f; cameraZ =  1.6f;
      sphLocX = 0.0f; sphLocY = 0.0f; sphLocZ = -1.0f;

      earthTexture = Utils.loadTexture("textures/earth.jpg");
   }

   private void setupVertices()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      mySphere = new Sphere(96);
      numSphereVerts = mySphere.getIndices().length;

      final int[] indices = mySphere.getIndices();
      final Vector3f[] vert = mySphere.getVertices();
      final Vector2f[] tex  = mySphere.getTexCoords();
      final Vector3f[] norm = mySphere.getNormals();

      final float[] pvalues = new float[indices.length * 3];
      final float[] tvalues = new float[indices.length * 2];
      final float[] nvalues = new float[indices.length * 3];

      for (int i = 0; i < indices.length; ++i)
      {
         pvalues[i*3]   = (float)(vert[indices[i]]).x;
         pvalues[i*3+1] = (float)(vert[indices[i]]).y;
         pvalues[i*3+2] = (float)(vert[indices[i]]).z;
         tvalues[i*2]   = (float)( tex[indices[i]]).x;
         tvalues[i*2+1] = (float)( tex[indices[i]]).y;
         nvalues[i*3]   = (float)(norm[indices[i]]).x;
         nvalues[i*3+1] = (float)(norm[indices[i]]).y;
         nvalues[i*3+2] = (float)(norm[indices[i]]).z;
      }

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(3, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer vertBuf = Buffers.newDirectFloatBuffer(pvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, vertBuf.limit()*4, vertBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      FloatBuffer texBuf = Buffers.newDirectFloatBuffer(tvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, texBuf.limit()*4, texBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
      FloatBuffer norBuf = Buffers.newDirectFloatBuffer(nvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, norBuf.limit()*4,norBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height)
   {
      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);
   }

   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_17();
   }
}
