/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 144-148.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.texture.*;

import org.joml.*;

import java.io.*;
import java.nio.*;
import java.lang.Math;
import javax.swing.*;

public class Code_18 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[4];
   private float cameraX, cameraY, cameraZ;
   private float torLocX, torLocY, torLocZ;
   private int renderingProgram;
   private int brickTexture;

   private Torus myTorus;
   private int numTorusIndices, numTorusVertices;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16);
   private final Matrix4f pMat = new Matrix4f();  // perspective matrix
   private final Matrix4f vMat = new Matrix4f();  // view matrix
   private final Matrix4f mMat = new Matrix4f();  // model matrix
   private final Matrix4f mvMat = new Matrix4f(); // model-view matrix
   private int mvLoc, projLoc;
   private float aspect;

   public Code_18()
   {
      this.setTitle("Chapter 6 - program 2  (Code_18)");
      this.setSize(800, 800);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClearColor(0.0f, 0.0f, 0.25f, 1.0f);
      gl.glClear(GL_DEPTH_BUFFER_BIT);

      gl.glUseProgram(renderingProgram);

        mvLoc = gl.glGetUniformLocation(renderingProgram, "mv_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      vMat.identity().setTranslation(-cameraX,-cameraY,-cameraZ);

      mMat.identity();
      mMat.translate(torLocX, torLocY, torLocZ);
      mMat.rotateX((float)Math.toRadians(30.0f));

      mvMat.identity();
      mvMat.mul(vMat);
      mvMat.mul(mMat);

      gl.glUniformMatrix4fv(mvLoc, 1, false, mvMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false, pMat.get(vals));

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      gl.glVertexAttribPointer(1, 2, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(1);

      gl.glActiveTexture(GL_TEXTURE0);
      gl.glBindTexture(GL_TEXTURE_2D, brickTexture);

      gl.glEnable(GL_CULL_FACE);
      gl.glFrontFace(GL_CCW);

      gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo[3]);
      gl.glDrawElements(GL_TRIANGLES, numTorusIndices, GL_UNSIGNED_INT, 0);
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_18.glsl",
                                                   "shaders/fragShader_18.glsl");

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);

      setupVertices();

      cameraX = 0.0f; cameraY = 0.0f; cameraZ =  1.2f;
      torLocX = 0.0f; torLocY = 0.0f; torLocZ = -0.5f;

      brickTexture = Utils.loadTexture("textures/brick1.jpg");
   }

   private void setupVertices()
   {  GL4 gl = (GL4) GLContext.getCurrentGL();

      myTorus = new Torus(0.5f, 0.2f, 48);
      numTorusVertices = myTorus.getNumVertices();
      numTorusIndices = myTorus.getNumIndices();

      Vector3f[] vertices = myTorus.getVertices();
      Vector2f[] texCoords = myTorus.getTexCoords();
      Vector3f[] normals = myTorus.getNormals();
      int[] indices = myTorus.getIndices();

      float[] pvalues = new float[vertices.length*3];
      float[] tvalues = new float[texCoords.length*2];
      float[] nvalues = new float[normals.length*3];

      for (int i=0; i<numTorusVertices; i++)
      {  pvalues[i*3]   = (float) vertices[i].x();
         pvalues[i*3+1] = (float) vertices[i].y();
         pvalues[i*3+2] = (float) vertices[i].z();
         tvalues[i*2]   = (float) texCoords[i].x();
         tvalues[i*2+1] = (float) texCoords[i].y();
         nvalues[i*3]   = (float) normals[i].x();
         nvalues[i*3+1] = (float) normals[i].y();
         nvalues[i*3+2] = (float) normals[i].z();
      }

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(4, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer vertBuf = Buffers.newDirectFloatBuffer(pvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, vertBuf.limit()*4, vertBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      FloatBuffer texBuf = Buffers.newDirectFloatBuffer(tvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, texBuf.limit()*4, texBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[2]);
      FloatBuffer norBuf = Buffers.newDirectFloatBuffer(nvalues);
      gl.glBufferData(GL_ARRAY_BUFFER, norBuf.limit()*4, norBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo[3]);
      IntBuffer idxBuf = Buffers.newDirectIntBuffer(indices);
      gl.glBufferData(GL_ELEMENT_ARRAY_BUFFER, idxBuf.limit()*4, idxBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h)
   {
      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);
   }

   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_18();
   }
}
