/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 22-24.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.opengl.glu.GLU;

import javax.swing.*;

public class Code_3 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private int renderingProgram;

   public Code_3()
   {
      this.setTitle("Chapter 2 - program 3 (Code_3)");
      this.setSize(600, 400);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }


   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) drawable.getGL();

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);

      // Arrays to collect GLSL compilation status values.
      // Note: One-element arrays are used because
      // the associated JOGL calls require arrays.
      final int[] vertCompiled = new int[1];
      final int[] fragCompiled = new int[1];
      final int[] programLinked = new int[1];

      final String vshaderSource[] =
      {
         "#version 430    \n",
         "void main(void) \n",
         "{ gl_Position = vec4(0.0, 0.0, 0.5, 1.0); } \n",
      };

      final String fshaderSource[] =
      {
         "#version 430    \n",
         "out vec4 color; \n",
         "void main(void) \n",
         "{ color = vec4(0.0, 0.0, 1.0, 1.0); } \n"
      };

      // Compile vertex shader, with error checking.
      final int vShader = gl.glCreateShader(GL_VERTEX_SHADER);
      gl.glShaderSource(vShader, vshaderSource.length, vshaderSource, null, 0);
      gl.glCompileShader(vShader);
      checkOpenGLError();  // can use returned boolean if desired
      gl.glGetShaderiv(vShader, GL_COMPILE_STATUS, vertCompiled, 0);
      if (vertCompiled[0] != 1)
      {
         System.out.println("vertex shader compilation failed");
         printShaderLog(vShader);
      }

      // Compile fragment shader, with error checking.
      final int fShader = gl.glCreateShader(GL_FRAGMENT_SHADER);
      gl.glShaderSource(fShader, fshaderSource.length, fshaderSource, null, 0);
      gl.glCompileShader(fShader);
      checkOpenGLError();  // can use returned boolean if desired
      gl.glGetShaderiv(fShader, GL_COMPILE_STATUS, fragCompiled, 0);
      if (fragCompiled[0] != 1)
      {
         System.out.println("fragment shader compilation failed");
         printShaderLog(fShader);
      }

      // Link shaders, with error checking.
      renderingProgram = gl.glCreateProgram();
      gl.glAttachShader(renderingProgram, vShader);
      gl.glAttachShader(renderingProgram, fShader);
      gl.glLinkProgram(renderingProgram);
      checkOpenGLError();
      gl.glGetProgramiv(renderingProgram, GL_LINK_STATUS, programLinked, 0);
      if (programLinked[0] != 1)
      {
         System.out.println("linking failed");
         printProgramLog(renderingProgram);
      }

      gl.glDeleteShader(vShader);
      gl.glDeleteShader(fShader);
   }


   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) drawable.getGL();

      gl.glUseProgram(renderingProgram);
      gl.glPointSize(5.0f);
      gl.glDrawArrays(GL_POINTS, 0, 1);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}


   private boolean checkOpenGLError()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();
      final GLU glu = new GLU();

      boolean foundError = false;
      int glErr = gl.glGetError();
      while (glErr != GL_NO_ERROR)
      {
         System.err.println("glError: " + glu.gluErrorString(glErr));
         foundError = true;
         glErr = gl.glGetError();
      }
      return foundError;
   }


   private void printShaderLog(int shader)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final int[] len = new int[1];
      final int[] chWrittn = new int[1];

      // determine the length of the shader compilation log
      gl.glGetShaderiv(shader, GL_INFO_LOG_LENGTH, len, 0);
      if (len[0] > 0)
      {
         final byte[] log = new byte[len[0]];
         gl.glGetShaderInfoLog(shader, len[0], chWrittn, 0, log, 0);
         System.out.println("Shader Info Log: ");
         for (int i = 0; i < log.length; ++i)
         {
            System.out.print((char) log[i]);
         }
      }
   }


   private void printProgramLog(int prog)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final int[] len = new int[1];
      final int[] chWrittn = new int[1];

      // determine length of the program compilation log
      gl.glGetProgramiv(prog, GL_INFO_LOG_LENGTH, len, 0);
      if (len[0] > 0)
      {
         final byte[] log = new byte[len[0]];
         gl.glGetProgramInfoLog(prog, len[0], chWrittn, 0, log, 0);
         System.out.println("Program Info Log: ");
         for (int i = 0; i < log.length; ++i)
         {
            System.out.print((char) log[i]);
         }
      }
   }


   public static void main(String[] args)
   {
      new Code_3();
   }
}